//
//  Visopsys
//  Copyright (C) 1998-2011 J. Andrew McLaughlin
// 
//  This program is free software; you can redistribute it and/or modify it
//  under the terms of the GNU General Public License as published by the Free
//  Software Foundation; either version 2 of the License, or (at your option)
//  any later version.
// 
//  This program is distributed in the hope that it will be useful, but
//  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
//  or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
//  for more details.
//  
//  You should have received a copy of the GNU General Public License along
//  with this program; if not, write to the Free Software Foundation, Inc.,
//  59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
//
//  kernelPciDriver.c
//

// These routines allow access to PCI configuration space.  Based on an
// original version contributed by Jonas Zaddach: See the file
// contrib/jonas-pci/src/kernel/kernelBusPCI.c

#include "kernelPciDriver.h"
#include "kernelBus.h"
#include "kernelDebug.h"
#include "kernelDevice.h"
#include "kernelError.h"
#include "kernelLog.h"
#include "kernelMalloc.h"
#include "kernelMisc.h"
#include "kernelProcessorX86.h"
#include <string.h>

#if defined(DEBUG)
#include "kernelPciCodes.h"
#endif

static pciSubClass subclass_old[] = {
  { 0x00, "other", DEVICECLASS_NONE, DEVICESUBCLASS_NONE },
  { 0x01, "VGA", DEVICECLASS_GRAPHIC, DEVICESUBCLASS_NONE },
  { PCI_INVALID_SUBCLASSCODE, "", DEVICECLASS_NONE, DEVICESUBCLASS_NONE }
};

static pciSubClass subclass_diskctrl[] = {
  { 0x00, "SCSI", DEVICECLASS_DISKCTRL, DEVICESUBCLASS_DISKCTRL_SCSI },
  { 0x01, "IDE", DEVICECLASS_DISKCTRL, DEVICESUBCLASS_DISKCTRL_IDE },
  { 0x02, "floppy", DEVICECLASS_DISKCTRL, DEVICESUBCLASS_NONE },
  { 0x03, "IPI", DEVICECLASS_DISKCTRL, DEVICESUBCLASS_NONE },
  { 0x04, "RAID", DEVICECLASS_DISKCTRL, DEVICESUBCLASS_NONE },
  { 0x05, "ATA", DEVICECLASS_DISKCTRL, DEVICESUBCLASS_NONE },
  { 0x06, "SATA", DEVICECLASS_DISKCTRL, DEVICESUBCLASS_DISKCTRL_SATA },
  { 0x80, "other", DEVICECLASS_DISKCTRL, DEVICESUBCLASS_NONE },
  { PCI_INVALID_SUBCLASSCODE, "", DEVICECLASS_NONE, DEVICESUBCLASS_NONE }
};

static pciSubClass subclass_net[] = {
  { 0x00, "ethernet", DEVICECLASS_NETWORK, DEVICESUBCLASS_NETWORK_ETHERNET },
  { 0x01, "token ring", DEVICECLASS_NETWORK, DEVICESUBCLASS_NONE },
  { 0x02, "FDDI", DEVICECLASS_NETWORK, DEVICESUBCLASS_NONE },
  { 0x03, "ATM", DEVICECLASS_NETWORK, DEVICESUBCLASS_NONE },
  { 0x04, "ISDN", DEVICECLASS_NETWORK, DEVICESUBCLASS_NONE },
  { 0x05, "WorldFip", DEVICECLASS_NETWORK, DEVICESUBCLASS_NONE },
  { 0x06, "PICMG", DEVICECLASS_NETWORK, DEVICESUBCLASS_NONE },
  { 0x80, "other", DEVICECLASS_NETWORK, DEVICESUBCLASS_NONE },
  { PCI_INVALID_SUBCLASSCODE, "", DEVICECLASS_NONE, DEVICESUBCLASS_NONE }
};

static pciSubClass subclass_graphics[] = {
  { 0x00, "VGA", DEVICECLASS_GRAPHIC, DEVICESUBCLASS_NONE },
  { 0x01, "XGA", DEVICECLASS_GRAPHIC, DEVICESUBCLASS_NONE },
  { 0x02, "3D", DEVICECLASS_GRAPHIC, DEVICESUBCLASS_NONE },
  { 0x80, "other", DEVICECLASS_GRAPHIC, DEVICESUBCLASS_NONE },
  { PCI_INVALID_SUBCLASSCODE, "", DEVICECLASS_NONE, DEVICESUBCLASS_NONE }
};

static pciSubClass subclass_multimed[] = {
  { 0x00, "video", DEVICECLASS_NONE, DEVICESUBCLASS_NONE },
  { 0x01, "audio", DEVICECLASS_NONE, DEVICESUBCLASS_NONE },
  { 0x02, "telephony", DEVICECLASS_NONE, DEVICESUBCLASS_NONE },
  { 0x03, "high-def audio", DEVICECLASS_NONE, DEVICESUBCLASS_NONE },
  { 0x80, "other", DEVICECLASS_NONE, DEVICESUBCLASS_NONE },
  { PCI_INVALID_SUBCLASSCODE, "", DEVICECLASS_NONE, DEVICESUBCLASS_NONE }
};

static pciSubClass subclass_mem[] = {
  { 0x00, "RAM", DEVICECLASS_MEMORY, DEVICESUBCLASS_NONE },
  { 0x01, "flash", DEVICECLASS_MEMORY, DEVICESUBCLASS_NONE },
  { 0x80, "other", DEVICECLASS_MEMORY, DEVICESUBCLASS_NONE },
  { PCI_INVALID_SUBCLASSCODE, "", DEVICECLASS_NONE, DEVICESUBCLASS_NONE }
};

static pciSubClass subclass_bridge[] = {
  { 0x00, "host", DEVICECLASS_NONE, DEVICESUBCLASS_NONE },
  { 0x01, "ISA", DEVICECLASS_NONE, DEVICESUBCLASS_NONE },
  { 0x02, "EISA", DEVICECLASS_NONE, DEVICESUBCLASS_NONE },
  { 0x03, "MCA", DEVICECLASS_NONE, DEVICESUBCLASS_NONE },
  { 0x04, "PCI/PCI", DEVICECLASS_NONE, DEVICESUBCLASS_NONE },
  { 0x05, "PCMCIA", DEVICECLASS_NONE, DEVICESUBCLASS_NONE },
  { 0x06, "nubus", DEVICECLASS_NONE, DEVICESUBCLASS_NONE },
  { 0x07, "cardbus", DEVICECLASS_NONE, DEVICESUBCLASS_NONE },
  { 0x08, "raceway", DEVICECLASS_NONE, DEVICESUBCLASS_NONE },
  { 0x09, "PCI/PCI", DEVICECLASS_NONE, DEVICESUBCLASS_NONE },
  { 0x0A, "infiniband", DEVICECLASS_NONE, DEVICESUBCLASS_NONE },
  { 0x80, "other", DEVICECLASS_NONE, DEVICESUBCLASS_NONE },
  { PCI_INVALID_SUBCLASSCODE, "", DEVICECLASS_NONE, DEVICESUBCLASS_NONE }
};

static pciSubClass subclass_comm[] = {
  { 0x00, "serial", DEVICECLASS_NONE, DEVICESUBCLASS_NONE },
  { 0x01, "parallel", DEVICECLASS_NONE, DEVICESUBCLASS_NONE },
  { 0x02, "multiport serial", DEVICECLASS_NONE, DEVICESUBCLASS_NONE },
  { 0x03, "modem", DEVICECLASS_NONE, DEVICESUBCLASS_NONE },
  { 0x04, "IEEE-488", DEVICECLASS_NONE, DEVICESUBCLASS_NONE },
  { 0x05, "smart card", DEVICECLASS_NONE, DEVICESUBCLASS_NONE },
  { 0x80, "other", DEVICECLASS_NONE, DEVICESUBCLASS_NONE },
  { PCI_INVALID_SUBCLASSCODE, "", DEVICECLASS_NONE, DEVICESUBCLASS_NONE }
};

static pciSubClass subclass_sys[] = {
  { 0x00, "(A)PIC", DEVICECLASS_PIC, DEVICESUBCLASS_NONE },
  { 0x01, "DMA", DEVICECLASS_DMA, DEVICESUBCLASS_NONE },
  { 0x02, "timer", DEVICECLASS_SYSTIMER, DEVICESUBCLASS_NONE },
  { 0x03, "RTC", DEVICECLASS_RTC, DEVICESUBCLASS_NONE },
  { 0x04, "PCI hotplug", DEVICECLASS_NONE, DEVICESUBCLASS_NONE },
  { 0x05, "SD controller", DEVICECLASS_NONE, DEVICESUBCLASS_NONE },
  { 0x80, "other", DEVICECLASS_NONE, DEVICESUBCLASS_NONE },
  { PCI_INVALID_SUBCLASSCODE, "", DEVICECLASS_NONE, DEVICESUBCLASS_NONE }
};

static pciSubClass subclass_input[] = {
  { 0x00, "keyboard", DEVICECLASS_KEYBOARD, DEVICESUBCLASS_NONE },
  { 0x01, "digitizer", DEVICECLASS_NONE, DEVICESUBCLASS_NONE },
  { 0x02, "mouse", DEVICECLASS_MOUSE, DEVICESUBCLASS_NONE },
  { 0x03, "scanner", DEVICECLASS_NONE, DEVICESUBCLASS_NONE },
  { 0x04, "gameport", DEVICECLASS_NONE, DEVICESUBCLASS_NONE },
  { 0x80, "other", DEVICECLASS_NONE, DEVICESUBCLASS_NONE },
  { PCI_INVALID_SUBCLASSCODE, "", DEVICECLASS_NONE, DEVICESUBCLASS_NONE }
};

static pciSubClass subclass_dock[] = {
  { 0x00, "generic", DEVICECLASS_NONE, DEVICESUBCLASS_NONE },
  { 0x80, "other", DEVICECLASS_NONE, DEVICESUBCLASS_NONE },
  { PCI_INVALID_SUBCLASSCODE, "", DEVICECLASS_NONE, DEVICESUBCLASS_NONE }
};

static pciSubClass subclass_cpu[] = {
  { 0x00, "386", DEVICECLASS_CPU, DEVICESUBCLASS_CPU_X86 },
  { 0x01, "486", DEVICECLASS_CPU, DEVICESUBCLASS_CPU_X86 },
  { 0x02, "pentium", DEVICECLASS_CPU, DEVICESUBCLASS_CPU_X86 },
  { 0x03, "P6", DEVICECLASS_CPU, DEVICESUBCLASS_CPU_X86 },
  { 0x10, "alpha", DEVICECLASS_CPU, DEVICESUBCLASS_NONE },
  { 0x20, "powerpc", DEVICECLASS_CPU, DEVICESUBCLASS_NONE },
  { 0x30, "MIPS", DEVICECLASS_CPU, DEVICESUBCLASS_NONE },
  { 0x40, "coprocessor", DEVICECLASS_CPU, DEVICESUBCLASS_NONE },
  { PCI_INVALID_SUBCLASSCODE, "", DEVICECLASS_NONE, DEVICESUBCLASS_NONE }
};

static pciSubClass subclass_serial[] = {
  { 0x00, "firewire", DEVICECLASS_BUS, DEVICESUBCLASS_NONE },
  { 0x01, "ACCESS.bus", DEVICECLASS_BUS, DEVICESUBCLASS_NONE },
  { 0x02, "SSA", DEVICECLASS_BUS, DEVICESUBCLASS_NONE },
  { 0x03, "USB", DEVICECLASS_BUS, DEVICESUBCLASS_BUS_USB },
  { 0x04, "fibre channel", DEVICECLASS_BUS, DEVICESUBCLASS_NONE },
  { 0x05, "smbus", DEVICECLASS_BUS, DEVICESUBCLASS_NONE },
  { 0x06, "infiniband", DEVICECLASS_BUS, DEVICESUBCLASS_NONE },
  { 0x07, "IPMI", DEVICECLASS_BUS, DEVICESUBCLASS_NONE },
  { 0x08, "sercos", DEVICECLASS_BUS, DEVICESUBCLASS_NONE },
  { 0x09, "canbus", DEVICECLASS_BUS, DEVICESUBCLASS_NONE },
  { PCI_INVALID_SUBCLASSCODE, "", DEVICECLASS_NONE, DEVICESUBCLASS_NONE }
};

static pciSubClass subclass_wireless[] = {
  { 0x00, "iRDA", DEVICECLASS_NETWORK, DEVICESUBCLASS_NETWORK_WIRELESS },
  { 0x01, "infrared", DEVICECLASS_NETWORK, DEVICESUBCLASS_NETWORK_WIRELESS },
  { 0x10, "radio", DEVICECLASS_NETWORK, DEVICESUBCLASS_NETWORK_WIRELESS },
  { 0x11, "bluetooth", DEVICECLASS_NETWORK, DEVICESUBCLASS_NETWORK_WIRELESS },
  { 0x12, "broadband", DEVICECLASS_NETWORK, DEVICESUBCLASS_NETWORK_WIRELESS },
  { 0x20, "802.11a ethernet", DEVICECLASS_NETWORK,
    DEVICESUBCLASS_NETWORK_WIRELESS },
  { 0x21, "802.11b ethernet", DEVICECLASS_NETWORK,
    DEVICESUBCLASS_NETWORK_WIRELESS },
  { 0x80, "other", DEVICECLASS_NETWORK, DEVICESUBCLASS_NETWORK_WIRELESS },
  { PCI_INVALID_SUBCLASSCODE, "", DEVICECLASS_NONE, DEVICESUBCLASS_NONE }
};

static pciSubClass subclass_intelio[] = {
  { 0x00, "I20/message", DEVICECLASS_NONE, DEVICESUBCLASS_NONE },
  { PCI_INVALID_SUBCLASSCODE, "", DEVICECLASS_NONE, DEVICESUBCLASS_NONE }
};

static pciSubClass subclass_sat[] = {
  { 0x01, "television", DEVICECLASS_NONE, DEVICESUBCLASS_NONE },
  { 0x02, "audio", DEVICECLASS_NONE, DEVICESUBCLASS_NONE },
  { 0x03, "voice", DEVICECLASS_NONE, DEVICESUBCLASS_NONE },
  { 0x04, "data", DEVICECLASS_NONE, DEVICESUBCLASS_NONE },
  { PCI_INVALID_SUBCLASSCODE, "", DEVICECLASS_NONE, DEVICESUBCLASS_NONE }
};

static pciSubClass subclass_encrypt[] = {
  { 0x00, "network", DEVICECLASS_NONE, DEVICESUBCLASS_NONE },
  { 0x10, "entertainment", DEVICECLASS_NONE, DEVICESUBCLASS_NONE },
  { 0x80, "other", DEVICECLASS_NONE, DEVICESUBCLASS_NONE },
  { PCI_INVALID_SUBCLASSCODE, "", DEVICECLASS_NONE, DEVICESUBCLASS_NONE }
};

static pciSubClass subclass_sigproc[] = {
  { 0x00, "DPIO", DEVICECLASS_NONE, DEVICESUBCLASS_NONE },
  { 0x01, "performance counter", DEVICECLASS_NONE, DEVICESUBCLASS_NONE },
  { 0x10, "communications synch", DEVICECLASS_NONE, DEVICESUBCLASS_NONE },
  { 0x20, "management", DEVICECLASS_NONE, DEVICESUBCLASS_NONE },
  { 0x80, "other", DEVICECLASS_NONE, DEVICESUBCLASS_NONE },
  { PCI_INVALID_SUBCLASSCODE, "", DEVICECLASS_NONE, DEVICESUBCLASS_NONE }
};

static pciSubClass subclass_prop[] = {
  { 0x80, "other", DEVICECLASS_NONE, DEVICESUBCLASS_NONE },
  { PCI_INVALID_SUBCLASSCODE, "", DEVICECLASS_NONE, DEVICESUBCLASS_NONE }
};

static pciClass pciClassNames[] = {
  { 0x00, "before PCI 2.0", subclass_old },
  { 0x01, "disk controller", subclass_diskctrl },
  { 0x02, "network interface", subclass_net },
  { 0x03, "graphics adapter", subclass_graphics },
  { 0x04, "multimedia controller", subclass_multimed },
  { 0x05, "memory controller", subclass_mem },
  { 0x06, "bridge device", subclass_bridge },
  { 0x07, "communication controller", subclass_comm },
  { 0x08, "system device", subclass_sys },
  { 0x09, "input device", subclass_input },
  { 0x0a, "docking station", subclass_dock },
  { 0x0b, "CPU", subclass_cpu },
  { 0x0c, "serial bus", subclass_serial },
  { 0x0d, "wireless controller", subclass_wireless },
  { 0x0e, "intelligent I/O controller", subclass_intelio },
  { 0x0f, "satellite controller", subclass_sat },
  { 0x10, "encryption controller", subclass_encrypt },
  { 0x11, "signal processing controller", subclass_sigproc },
  { 0xFF, "proprietary device", subclass_prop },
  { PCI_INVALID_CLASSCODE, "", NULL }
};

static const char *unknownDevice = "unknown";
static const char *otherDevice = "other";

static kernelBusTarget *targets = NULL;
static int numTargets = 0;

#define headerAddress(bus, device, function, reg)                       \
  (0x80000000L |  (((unsigned) ((bus) & 0xFF) << 16) |                  \
   (((device) & 0x1F)  << 11) | (((function) & 0x07) << 8) |            \
   (((reg) & 0x3F) << 2)))

// Make our proprietary PCI target code
#define makeTargetCode(bus, device, function)                           \
  ((((bus) & 0xFF) << 16) | (((device) & 0xFF) << 8) | ((function) & 0xFF))

// Translate a target code back to bus, device, function
#define makeBusDevFunc(targetCode, bus, device, function)               \
  {  (bus) = (((targetCode) >> 16) & 0xFF);                             \
     (device) = (((targetCode) >> 8) & 0xFF);                           \
     (function) = ((targetCode) & 0xFF);  }


static void readConfig8(int bus, int dev, int function, int reg,
			unsigned char *data)
{
  // Reads configuration byte
  unsigned address = headerAddress(bus, dev, function, (reg / 4));
  kernelProcessorOutPort32(PCI_CONFIG_PORT, address);
  kernelProcessorInPort8((PCI_DATA_PORT + (reg % 4)), *data);
  return;
}


static void writeConfig8(int bus, int dev, int function, int reg,
			 unsigned char data)
{
  // Writes configuration byte
  unsigned address = headerAddress(bus, dev, function, (reg / 4));
  kernelProcessorOutPort32(PCI_CONFIG_PORT, address);
  kernelProcessorOutPort8((PCI_DATA_PORT + (reg % 4)), data);
  return;
}


static void readConfig16(int bus, int dev, int function, int reg,
			 unsigned short *data)
{
  // Reads configuration word
  unsigned address = headerAddress(bus, dev, function, (reg / 2));
  kernelProcessorOutPort32(PCI_CONFIG_PORT, address);
  kernelProcessorInPort16((PCI_DATA_PORT + (reg % 2)), *data);
  return;
}


static void writeConfig16(int bus, int dev, int function, int reg,
			  unsigned short data)
{
  // Writes configuration word
  unsigned address = headerAddress(bus, dev, function, (reg / 2));
  kernelProcessorOutPort32(PCI_CONFIG_PORT, address);
  kernelProcessorOutPort16((PCI_DATA_PORT + (reg % 2)), data);
  return;
}


static void readConfig32(int bus, int dev, int function, int reg,
			 unsigned *data)
{
  // Reads configuration dword
  unsigned address = headerAddress(bus, dev, function, reg);
  kernelProcessorOutPort32(PCI_CONFIG_PORT, address);
  kernelProcessorInPort32(PCI_DATA_PORT, *data);
  return;
}


static void writeConfig32(int bus, int dev, int function, int reg,
			  unsigned data)
{
  // Writes configuration dword
  unsigned address = headerAddress(bus, dev, function, reg);
  kernelProcessorOutPort32(PCI_CONFIG_PORT, address);
  kernelProcessorOutPort32(PCI_DATA_PORT, data);
  return;
}


static void readConfigHeader(int bus, int dev, int function,
			     pciDeviceInfo *devInfo)
{
  // Fill up the supplied device info header

  unsigned address = 0;
  int count;

  for (count = 0; count < PCI_CONFIGHEADER_SIZE; count ++)
    {
      address =	headerAddress(bus, dev, function, count);
      kernelProcessorOutPort32(PCI_CONFIG_PORT, address);
      kernelProcessorInPort32(PCI_DATA_PORT, devInfo->header[count]);
    }
}


static void getClass(int classCode, pciClass **class)
{
  // Return the PCI class, given the class code

  int count;

  for (count = 0; count < 256; count++)
    {	
      // If no more classcodes are in the list
      if (pciClassNames[count].classCode == PCI_INVALID_CLASSCODE)
	{
	  *class = NULL;
	  return;
	}
		
      // If valid classcode is found
      if (pciClassNames[count].classCode == classCode)
	{
	  *class = &(pciClassNames[count]);
	  return;
	}
    }
}


static void getSubClass(pciClass *class, int subClassCode,
			pciSubClass **subClass)
{
  // Return the PCI subclass, given the class and subclass code

  int count;

  for (count = 0; count < 256; count++)
    {	
      // If no more subclass codes are in the list
      if (class->subClasses[count].subClassCode == PCI_INVALID_SUBCLASSCODE)
	{
	  *subClass = NULL;
	  return;
	}
		
      // If valid subclass code is found
      if (class->subClasses[count].subClassCode == subClassCode)
	{
	  *subClass = &(class->subClasses[count]);
	  return;
	}
    }
}


static int getClassName(int classCode, int subClassCode, char **className,
			char **subClassName)
{
  // Returns name of the class and the subclass in human readable format.
  // Buffers classname and subclassname have to provide

  int status = 0;
  pciClass *class = NULL;
  pciSubClass *subClass = NULL;

  getClass(classCode, &class);
  if (class == NULL)
    {
      *className = (char *) unknownDevice;
      return (status = PCI_INVALID_CLASSCODE);
    }

  *className = (char *) class->name;

  // Subclasscode 0x80 is always other
  if (subClassCode == 0x80)
    {
      *subClassName = (char *) otherDevice;
      return (status = 0);
    }

  getSubClass(class, subClassCode, &subClass);
  if (subClass == NULL)
    {
      *subClassName = (char *) unknownDevice;
      return (status = PCI_INVALID_SUBCLASSCODE);
    }

  *subClassName = (char *) subClass->name;
  return (status = 0);
}


#if defined(DEBUG)
static const char *getVendorFromCode(unsigned short vendorCode
				     __attribute__((unused)),
				     int longDesc __attribute__((unused)))
{
  const char *vendor = "unknown";
  unsigned count;

  for (count = 0; count < PCI_VENTABLE_LEN; count ++)
    if (PciVenTable[count].VenId == vendorCode)
      {
	if (longDesc)
	  vendor = PciVenTable[count].VenFull;
	else
	  vendor = PciVenTable[count].VenShort;

	break;
      }

  return (vendor);
}


static const char *getDeviceFromCode(unsigned short vendorCode
				     __attribute__((unused)),
				     unsigned short deviceCode
				     __attribute__((unused)),
				     int longDesc __attribute__((unused)))
{
  const char *dev = "unknown";
  unsigned count;

  for (count = 0; count < PCI_DEVTABLE_LEN; count ++)
    if ((PciDevTable[count].VenId == vendorCode) &&
	(PciDevTable[count].DevId == deviceCode))
      {
	if (longDesc)
	  dev = PciDevTable[count].ChipDesc;
	else
	  dev = PciDevTable[count].Chip;

	break;
      }

  return (dev);
}
#endif


static void deviceInfo2BusTarget(kernelBus *bus, int busNum, int dev,
				 int function, pciDeviceInfo *info,
				 kernelBusTarget *target)
{
  // Translate a device info header to a bus target listing

  pciClass *class = NULL;
  pciSubClass *subClass = NULL;

  target->bus = bus;
  target->id = makeTargetCode(busNum, dev, function);

  getClass(info->device.classCode, &class);
  if (class == NULL)
    {
      kernelDebugError("No class for classCode %02x", info->device.classCode);
      return;
    }

  getSubClass(class, info->device.subClassCode, &subClass);
  if (subClass == NULL)
    {
      kernelDebugError("No subclass for classCode %02x, subClassCode %02x",
		       info->device.classCode, info->device.subClassCode);
      return;
    }

  target->class = kernelDeviceGetClass(subClass->systemClassCode);
  target->subClass = kernelDeviceGetClass(subClass->systemSubClassCode);
}


static int driverGetTargets(kernelBus *bus, kernelBusTarget **pointer)
{
  // Generate the list of targets that reside on the given bus (controller).

  int status = 0;
  int targetCount = 0;
  int count;

  // Count up the number of targets that belong to the bus
  for (count = 0; count < numTargets; count ++)
    if (targets[count].bus == bus)
      targetCount += 1;

  if (!targetCount)
    return (targetCount = 0);

  // Allocate memory for all the targets
  *pointer = kernelMalloc(targetCount * sizeof(kernelBusTarget));
  if (*pointer == NULL)
    return (status = ERR_MEMORY);

  // Loop again and copy targets
  for (targetCount = 0, count = 0; count < numTargets; count ++)
    if (targets[count].bus == bus)
      {
	kernelMemCopy(&targets[count], &(*pointer)[targetCount],
		      sizeof(kernelBusTarget));
	targetCount += 1;
      }

  return (targetCount);
}


static int driverGetTargetInfo(kernelBusTarget *target, void *pointer)
{
  // Read the device's PCI header and copy it to the supplied memory
  // pointer

  int status = 0;
  int bus, dev, function;

  makeBusDevFunc(target->id, bus, dev, function);
  readConfigHeader(bus, dev, function, pointer);

  return (status = 0);
}


static unsigned driverReadRegister(kernelBusTarget *target, int reg,
				   int bitWidth)
{
  // Returns the contents of a PCI configuration register

  unsigned contents = 0;
  int bus, dev, function;

  makeBusDevFunc(target->id, bus, dev, function);

  switch (bitWidth)
    {
    case 8:
      readConfig8(bus, dev, function, reg, (unsigned char *) &contents);
      break;
    case 16:
      readConfig16(bus, dev, function, reg, (unsigned short *) &contents);
      break;
    case 32:
      readConfig32(bus, dev, function, reg, &contents);
      break;
    default:
      kernelError(kernel_error, "Register width %d not supported", bitWidth);
    }

  return (contents);
}


static int driverWriteRegister(kernelBusTarget *target, int reg, int bitWidth,
			       unsigned contents)
{
  // Write the contents of a PCI configuration register

  int status = 0;
  int bus, dev, function;
  
  makeBusDevFunc(target->id, bus, dev, function);

  switch (bitWidth)
    {
    case 8:
      writeConfig8(bus, dev, function, reg, (unsigned char) contents);
      break;
    case 16:
      writeConfig16(bus, dev, function, reg, (unsigned short) contents);
      break;
    case 32:
      writeConfig32(bus, dev, function, reg, contents);
      break;
    default:
      kernelError(kernel_error, "Register width %d not supported", bitWidth);
      status = ERR_RANGE;
      break;
    }

  return (status);
}


static void driverDeviceClaim(kernelBusTarget *target, kernelDriver *driver)
{
  // Allows a driver to claim a PCI bus device

  int count;

  // Find our copy of the target using the ID
  for (count = 0; count < numTargets; count ++)
    if (targets[count].id == target->id)
      {
	kernelDebug(debug_pci, "PCI: target %08x claimed", targets[count].id);
	targets[count].claimed = driver;
      }

  return;
}


static int driverDeviceEnable(kernelBusTarget *target, int enable)
{
  // Enables or disables a PCI bus device

  int bus, dev, function;
  unsigned short commandReg = 0;

  makeBusDevFunc(target->id, bus, dev, function);

  // Read the command register
  readConfig16(bus, dev, function, PCI_CONFREG_COMMAND_16, &commandReg);
  
  if (enable)
    {
      if (enable & PCI_COMMAND_IOENABLE)
	// Turn on I/O access
	commandReg |= PCI_COMMAND_IOENABLE;
      if (enable & PCI_COMMAND_MEMORYENABLE)
	// Turn on memory access
	commandReg |= PCI_COMMAND_MEMORYENABLE;
    }
  else
    // Turn off I/O access and memory access
    commandReg &= ~(PCI_COMMAND_IOENABLE | PCI_COMMAND_MEMORYENABLE);

  // Write back command register
  writeConfig16(bus, dev, function, PCI_CONFREG_COMMAND_16, commandReg);

  return (0);
}


static int driverSetMaster(kernelBusTarget *target, int master)
{
  // Sets the target device as a bus master

  int bus, dev, function;
  unsigned short commandReg = 0;
  unsigned char latency = 0;

  makeBusDevFunc(target->id, bus, dev, function);

  // Read the command register
  readConfig16(bus, dev, function, PCI_CONFREG_COMMAND_16, &commandReg);

  // Toggle busmaster bit
  if (master)
    commandReg |= PCI_COMMAND_MASTERENABLE;
  else
    commandReg &= ~PCI_COMMAND_MASTERENABLE;

  // Write back command register
  writeConfig16(bus, dev, function, PCI_CONFREG_COMMAND_16, commandReg);

  // Check latency timer
  readConfig8(bus, dev, function, PCI_CONFREG_LATENCY_8, &latency);

  if (latency < 0x10)
    {
      latency = 0x40;
      writeConfig8(bus, dev, function, PCI_CONFREG_LATENCY_8, latency);
    }

  return (0);
}


static int driverDetect(void *parent, kernelDriver *driver)
{
  // This routine is used to detect and initialize each device, as well as
  // registering each one with any higher-level interfaces.

  int status = 0;
  unsigned reply = 0;
  int busCount = 0;
  int deviceCount = 0;
  int functionCount = 0;
  char *className = NULL;
  char *subclassName = NULL;
  pciDeviceInfo pciDevice;
  kernelDevice *dev = NULL;
  kernelBus *bus = NULL;

  // Check for a configuration mechanism #1 able PCI controller.
  kernelProcessorOutPort32(PCI_CONFIG_PORT, 0x80000000L);
  kernelProcessorInPort32(PCI_CONFIG_PORT, reply);

  if (reply != 0x80000000L)
    // No device that uses configuration mechanism #1.  Fine enough: No PCI
    // functionality for you.
    return (status = 0);

  // First count all the devices on the bus
  for (busCount = 0; busCount < PCI_MAX_BUSES; busCount ++) 
    for (deviceCount = 0; deviceCount < PCI_MAX_DEVICES; deviceCount ++)
      for (functionCount = 0; functionCount < PCI_MAX_FUNCTIONS;
	   functionCount ++) 
	{
	  // Just read the first dword of the header to get the device and
	  // vendor IDs
	  readConfig32(busCount, deviceCount, functionCount, 0,
		       &(pciDevice.header[0]));

	  // See if this is really a device, or if this device header is
	  // unoccupied.
	  if ((pciDevice.device.vendorID == 0x0000) ||
	      (pciDevice.device.vendorID == 0xffff) ||
	      (pciDevice.device.deviceID == 0xffff))
	    // No device here, so try next one
	    continue;

	  kernelDebug(debug_pci, "PCI: found device bus=%d device=%d "
		      "function=%d", busCount, deviceCount, functionCount);

	  // If here, we found a PCI device
	  numTargets += 1;
	}

  // Allocate memory for the PCI bus device
  dev = kernelMalloc(sizeof(kernelDevice));
  if (dev == NULL)
    return (status = ERR_MEMORY);

  dev->device.class = kernelDeviceGetClass(DEVICECLASS_BUS);
  dev->device.subClass = kernelDeviceGetClass(DEVICESUBCLASS_BUS_PCI);
  dev->driver = driver;

  // Allocate memory for the bus service
  bus = kernelMalloc(sizeof(kernelBus));
  if (bus == NULL)
    return (status = ERR_MEMORY);

  bus->type = bus_pci;
  bus->dev = dev;
  bus->ops = driver->ops;

  // Allocate memory for the targets list
  targets = kernelMalloc(numTargets * sizeof(kernelBusTarget));
  if (targets == NULL)
    return (status = ERR_MEMORY);

  // Now fill up our targets list
  for (numTargets = 0, busCount = 0; busCount < PCI_MAX_BUSES; busCount ++) 
    for (deviceCount = 0; deviceCount < PCI_MAX_DEVICES; deviceCount ++)
      for (functionCount = 0; functionCount < PCI_MAX_FUNCTIONS;
	   functionCount ++) 
	{
	  // Just read the first dword of the header to get the device and
	  // vendor IDs
	  readConfig32(busCount, deviceCount, functionCount, 0,
		       &(pciDevice.header[0]));

	  if ((pciDevice.device.vendorID == 0x0000) ||
	      (pciDevice.device.vendorID == 0xffff) ||
	      (pciDevice.device.deviceID == 0xffff))
	    // No device here, so try next one
	    continue;

	  // There's a device.  Get the full device header.
	  readConfigHeader(busCount, deviceCount, functionCount, &pciDevice);

	  getClassName(pciDevice.device.classCode,
		       pciDevice.device.subClassCode, &className,
		       &subclassName);

	  kernelLog("PCI: %s %s %u:%u:%u vend:%04x, dev:%04x, class:%02x, "
		    "sub:%02x", subclassName, className, busCount,
		    deviceCount, functionCount, pciDevice.device.vendorID,
		    pciDevice.device.deviceID, pciDevice.device.classCode,
		    pciDevice.device.subClassCode);

	  kernelDebug(debug_pci, "PCI: vendor=\"%s\" device=\"%s\"",
		      getVendorFromCode(pciDevice.device.vendorID, 1),
		      getDeviceFromCode(pciDevice.device.vendorID,
					pciDevice.device.deviceID, 1));

	  deviceInfo2BusTarget(bus, busCount, deviceCount, functionCount,
			       &pciDevice, &targets[numTargets]);
	  numTargets += 1;
	}

  // Register the device
  status = kernelDeviceAdd(parent, dev);
  if (status < 0)
    {
      kernelFree(bus);
      kernelFree(dev);
      return (status);
    }

  // Register the bus service
  status = kernelBusRegister(bus);
  if (status < 0)
    {
      kernelFree(bus);
      kernelFree(dev);
      return (status);
    }

  return (status = 0);
}


// Our driver operations structure.
static kernelBusOps pciOps = {
  driverGetTargets,
  driverGetTargetInfo,
  driverReadRegister,
  driverWriteRegister,
  driverDeviceClaim,
  driverDeviceEnable,
  driverSetMaster,
  NULL, // driverRead
  NULL  // driverWrite
};


/////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////
//
//  Below here, the functions are exported for external use
//
/////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////


void kernelPciDriverRegister(kernelDriver *driver)
{
   // Device driver registration.

  driver->driverDetect = driverDetect;
  driver->ops = &pciOps;

  return;
}
